from dataclasses import dataclass
from typing import Type, Optional, Union

import typing_inspect

from marshy.marshy_context import MarshyContext
from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.marshaller import marshaller_abc
from marshy.marshaller.deferred_marshaller import DeferredMarshaller
from marshy.marshaller.union_marshaller import UnionMarshaller


@dataclass
class UnionMarshallerFactory(MarshallerFactoryABC):
    priority: int = 90

    def create(
        self, context: MarshyContext, type_: Type
    ) -> Optional[marshaller_abc.MarshallerABC]:
        origin = typing_inspect.get_origin(type_)
        if origin == Union:
            marshallers = {
                name_for_type(t): DeferredMarshaller(t, context)
                for t in typing_inspect.get_args(type_)
            }
            return UnionMarshaller[type_](type_, marshallers)


def name_for_type(type_: Type) -> str:
    if hasattr(type_, "__origin__"):
        return name_for_type(type_.__origin__)
    return type_.__name__ if hasattr(type_, "__name__") else str(type_)



from datetime import datetime
from typing import Type

from marshy.marshaller.marshaller_abc import MarshallerABC


class DatetimeMarshaller(MarshallerABC[datetime]):
    """
    Marshaller for datetime instances to iso format
    """

    marshalled_type: Type[datetime] = datetime

    def load(self, item: str) -> datetime:
        loaded = datetime.fromisoformat(item)
        return loaded

    def dump(self, item: datetime) -> str:
        dumped = item.isoformat()
        return dumped



from dataclasses import dataclass
from typing import TypeVar, Iterable, List, Callable, Iterator

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC

T = TypeVar("T")


@dataclass(frozen=True)
class IterableMarshaller(MarshallerABC[Iterable[T]]):
    """
    Marshaller for iterable types (lists)
    """

    marshalled_type: T
    item_marshaller: MarshallerABC[T]
    constructor: Callable[[Iterator[T]], Iterable[T]] = list

    def load(self, item: List[ExternalType]) -> Iterable[T]:
        loaded = self.constructor(self.item_marshaller.load(i) for i in item)
        return loaded

    def dump(self, item: Iterable[T]) -> List[ExternalType]:
        dumped = [self.item_marshaller.dump(i) for i in item]
        return dumped



from dataclasses import dataclass, field
from typing import Type, Optional

from injecty import get_default_injecty_context, InjectyContext

from marshy.marshy_context import MarshyContext
from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.factory.union_marshaller_factory import name_for_type
from marshy.marshaller import marshaller_abc
from marshy.marshaller.deferred_marshaller import DeferredMarshaller
from marshy.marshaller.union_marshaller import UnionMarshaller

MARSHALLER_FACTORY = "__marshaller_factory__"


@dataclass(frozen=True)
class ImplMarshallerFactory(MarshallerFactoryABC):
    injecty_context: InjectyContext = field(default_factory=get_default_injecty_context)
    priority: int = 110

    def create(
        self, context: MarshyContext, type_: Type
    ) -> Optional[marshaller_abc.MarshallerABC]:
        # noinspection PyTypeChecker
        impls = self.injecty_context.get_impls(base=type_, permit_no_impl=True)
        if impls:
            marshallers = {
                name_for_type(i): DeferredMarshaller[i](i, context) for i in impls
            }
            return UnionMarshaller[type_](type_, marshallers)



from dataclasses import dataclass
from types import NoneType
from typing import Type, Optional, Union

import typing_inspect

from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.marshaller import marshaller_abc
from marshy.marshaller.optional_marshaller import OptionalMarshaller
from marshy.marshy_context import MarshyContext


@dataclass
class OptionalMarshallerFactory(MarshallerFactoryABC):
    priority: int = 100

    def create(
        self, context: MarshyContext, type_: Type
    ) -> Optional[marshaller_abc.MarshallerABC]:
        origin = typing_inspect.get_origin(type_)
        if origin == Union:
            optional_type = get_optional_type(type_)
            if optional_type:
                return OptionalMarshaller[optional_type](
                    context.get_marshaller(optional_type)
                )


def get_optional_type(type_: Type) -> Optional[Type]:
    origin = typing_inspect.get_origin(type_)
    if origin == Union:
        args = typing_inspect.get_args(type_)
        if len(args) != 2:
            return None
        optional_args = [a for a in args if a != NoneType]
        if len(optional_args) == 1:
            return optional_args[0]



from typing import TypeVar, Optional

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC

T = TypeVar("T")


class OptionalMarshaller(MarshallerABC[Optional[T]]):
    """
    Marshaller for optional types
    """

    def __init__(self, marshaller: MarshallerABC[T]):
        self.marshalled_type = Optional[marshaller.marshalled_type]
        # noinspection PyDataclass
        self.marshaller = marshaller

    def load(self, item: ExternalType) -> Optional[T]:
        if item is not None:
            loaded = self.marshaller.load(item)
            return loaded

    def dump(self, item: Optional[T]) -> ExternalType:
        if item is not None:
            dumped = self.marshaller.dump(item)
            return dumped



from typing import Type

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC


class BoolMarshaller(MarshallerABC[ExternalType]):
    marshalled_type: Type[bool] = bool

    def load(self, item: ExternalType) -> bool:
        if isinstance(item, str):
            return item.lower() not in ["", "0", "0.0", "false"]  # Custom falsy values
        return bool(item)

    def dump(self, item: bool) -> bool:
        return bool(item)



import json
from typing import TypeVar, Type, Optional

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType

T = TypeVar("T")


class JsonStrMarshaller(MarshallerABC[T]):
    """
    Marshaller for freeform json objects - they are interpreted as strings. (Due to technologies like graphql
    having a dislike for unstructured data)
    """

    def load(self, item: str) -> T:
        loaded = json.loads(item)
        return loaded

    def dump(self, item: T) -> str:
        dumped = json.dumps(item)
        return dumped


class ExternalTypeMarshaller(JsonStrMarshaller):
    marshalled_type: Type[ExternalType] = ExternalType


class ExternalItemTypeMarshaller(JsonStrMarshaller):
    marshalled_type: Type[ExternalItemType] = ExternalItemType


class OptionalExternalItemTypeMarshaller(JsonStrMarshaller):
    marshalled_type: Type[Optional[ExternalItemType]] = Optional[ExternalItemType]



from dataclasses import dataclass
from typing import TypeVar

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.marshy_context import MarshyContext

T = TypeVar("T")


@dataclass(frozen=True)
class DeferredMarshaller(MarshallerABC[T]):
    marshalled_type: T
    marshaller_context: MarshyContext

    def get_marshaller(self) -> MarshallerABC[T]:
        marshaller = self.marshaller_context.get_marshaller(self.marshalled_type)
        return marshaller

    def load(self, item: ExternalType) -> T:
        loaded = self.get_marshaller().load(item)
        return loaded

    def dump(self, item: T) -> ExternalType:
        dumped = self.get_marshaller().dump(item)
        return dumped



import importlib
from dataclasses import dataclass, field
from typing import Type, Iterable, Dict

from marshy.errors import MarshallError
from marshy.marshaller.marshaller_abc import MarshallerABC


@dataclass(frozen=True)
class TypeMarshaller(MarshallerABC[Type]):
    """
    Marshaller for object types - allows setting up a preset 'safe' module list
    """

    marshalled_type: Type = Type
    permitted_prefixes: Iterable[str] = field(default_factory=tuple)
    _names_to_types: Dict[str, Type] = field(default_factory=dict)
    _types_to_names: Dict[Type, str] = None

    def __post_init__(self):
        types_to_names = {t: n for n, t in self._names_to_types.items()}
        object.__setattr__(self, "_types_to_names", types_to_names)

    def load(self, item: str) -> Type:
        type_ = self._names_to_types.get(item)
        if type_ is not None:
            return type_
        for permitted_prefix in self.permitted_prefixes:
            if item.startswith(permitted_prefix):
                path = item.split(".")
                module = ".".join(path[:-1])
                module = importlib.import_module(module)
                type_ = getattr(module, path[-1])
                self._names_to_types[item] = type_
                self._types_to_names[type_] = item
                return type_
        raise MarshallError(f"type_not_permitted:{item}")

    def dump(self, item: Type) -> str:
        name = self._types_to_names.get(item)
        if name is not None:
            return name
        path = item.__module__ + "." + item.__name__
        for permitted_prefix in self.permitted_prefixes:
            if path.startswith(permitted_prefix):
                self._names_to_types[path] = item
                self._types_to_names[item] = path
                return path
        raise MarshallError(f"type_not_permitted:{item}")



from dataclasses import dataclass, field
from typing import Type, Any, Optional, Dict, List

from injecty import InjectyContext, get_default_injecty_context

from marshy.errors import MarshallError
from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.utils import resolve_forward_refs
from marshy.types import ExternalType


@dataclass
class MarshyContext:
    marshallers_by_type: Dict[Type, MarshallerABC] = field(default_factory=dict)
    factories: List[MarshallerFactoryABC] = field(default_factory=list)

    def load(self, type_: Type, to_load: ExternalType):
        marshaller = self.get_marshaller(type_)
        loaded = marshaller.load(to_load)
        return loaded

    def dump(self, obj: Any, type_: Optional[Type] = None) -> ExternalType:
        if not type_:
            type_ = type(obj)
        marshaller = self.get_marshaller(type_)
        dumped = marshaller.dump(obj)
        return dumped

    def get_marshaller(self, type_: Type) -> MarshallerABC:
        marshaller = self.marshallers_by_type.get(type_)
        if not marshaller:
            resolved_type = resolve_forward_refs(type_)
            for factory in self.factories:
                marshaller = factory.create(self, resolved_type)
                if marshaller:
                    break
            if not marshaller:
                raise MarshallError(f"no_marshaller_for_type:{type_}")
            self.marshallers_by_type[type_] = marshaller
        return marshaller

    def register_marshaller(self, marshaller: MarshallerABC):
        self.marshallers_by_type[marshaller.marshalled_type] = marshaller

    def register_factory(self, factory: MarshallerFactoryABC):
        self.factories.append(factory)
        self.factories.sort(key=lambda f: f.priority, reverse=True)


def marshy_context(injecty_context: Optional[InjectyContext] = None):
    if injecty_context is None:
        injecty_context = get_default_injecty_context()
    return MarshyContext(
        marshallers_by_type={
            m.marshalled_type: m for m in injecty_context.get_instances(MarshallerABC)
        },
        factories=injecty_context.get_instances(MarshallerFactoryABC),
    )



from typing import Type

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC


class PrimitiveMarshaller(MarshallerABC[ExternalType]):
    """
    Marshaller for cases where the item is already an external type (int, float, str)
    """

    def load(self, item: ExternalType) -> ExternalType:
        return None if item is None else self.marshalled_type(item)

    def dump(self, item: ExternalType) -> ExternalType:
        return item


class FloatMarshaller(PrimitiveMarshaller):
    marshalled_type: Type[float] = float


class IntMarshaller(PrimitiveMarshaller):
    marshalled_type: Type[int] = int


class StrMarshaller(PrimitiveMarshaller):
    marshalled_type: Type[str] = str



from typing import Type
from uuid import UUID

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC


class AsStrMarshaller(MarshallerABC[ExternalType]):
    """
    Marshaller for cases where item is easily converted back and forth from strings
    """

    def load(self, item: ExternalType) -> ExternalType:
        return None if item is None else self.marshalled_type(item)

    def dump(self, item: ExternalType) -> ExternalType:
        return str(item)


class UuidMarshaller(AsStrMarshaller):
    marshalled_type: Type[UUID] = UUID



from dataclasses import dataclass, field
from typing import TypeVar, Optional, Dict, Type, List, Iterable

from marshy import ExternalType
from marshy.marshaller.deferred_marshaller import DeferredMarshaller
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.marshy_context import MarshyContext

T = TypeVar("T")


@dataclass(frozen=True)
class UnionMarshaller(MarshallerABC[T]):
    """
    Marshaller for polymorphic types
    """

    marshalled_type: T
    marshallers_by_name: Dict[str, MarshallerABC[T]]
    names_by_type: Dict[Type, str] = field(default_factory=dict)

    def __post_init__(self):
        self.names_by_type.update(
            {
                resolve_type(m.marshalled_type): n
                for n, m in self.marshallers_by_name.items()
            }
        )

    def load(self, item: List[ExternalType]) -> T:
        type_name = item[0]
        marshaller = self.marshallers_by_name[type_name]
        loaded = marshaller.load(item[1])
        return loaded

    def dump(self, item: Optional[T]) -> List[ExternalType]:
        type_name = self.names_by_type[item.__class__]
        marshaller = self.marshallers_by_name[type_name]
        dumped = marshaller.dump(item)
        return [type_name, dumped]


def resolve_type(type_: Type) -> Type:
    return resolve_type(type_.__origin__) if hasattr(type_, "__origin__") else type_


def implementation_marshaller(
    base_type: Type, impls: Iterable[Type], marshaller_context: MarshyContext
):
    return UnionMarshaller(
        base_type,
        {i.__name__: DeferredMarshaller[i](i, marshaller_context) for i in impls},
    )



from types import NoneType

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC


class NoOpMarshaller(MarshallerABC[ExternalType]):
    """
    Marshaller for NoOps (Mainly None)
    """

    def load(self, item: ExternalType) -> ExternalType:
        return item

    def dump(self, item: ExternalType) -> ExternalType:
        return item


class NoneMarshaller(NoOpMarshaller):
    marshalled_type: NoneType = NoneType



from abc import ABC, abstractmethod
from functools import total_ordering
from typing import Optional, Type

_MarshyContext = "marshy.marshy_context.MarshyContext"
_MarshallerABC = "marshy.marshaller_abc.MarshallerABC"


@total_ordering
class MarshallerFactoryABC(ABC):
    priority: int = 0

    @abstractmethod
    def create(self, context: _MarshyContext, type_: Type) -> Optional[_MarshallerABC]:
        """
        Create a new marshaller instance if possible - if not return None
        """

    def __ne__(self, other):
        return self.priority != getattr(other, "priority", None)

    def __lt__(self, other):
        return self.priority < getattr(other, "priority", None)



from dataclasses import dataclass
from typing import TypeVar, Tuple, Optional, Type

from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalType

T = TypeVar("T")


@dataclass(frozen=True)
class AttrConfig:
    internal_name: str
    external_name: str
    marshaller: MarshallerABC
    load: bool
    dump: bool
    exclude_dumped_values: Tuple = ()


# pylint: disable=R0913
def attr_config(
    marshaller: MarshallerABC,
    internal_name: str,
    external_name: Optional[str] = None,
    load: bool = True,
    dump: bool = True,
    exclude_dumped_values: Tuple = tuple(),
) -> AttrConfig:
    if external_name is None:
        external_name = internal_name
    return AttrConfig(
        internal_name, external_name, marshaller, load, dump, exclude_dumped_values
    )


@dataclass(frozen=True)
class ObjMarshaller(MarshallerABC[T]):
    marshalled_type: Type[T]
    attr_configs: Tuple

    def load(self, item: ExternalType) -> T:
        kwargs = {}
        for a in self.attr_configs:
            if a.load and a.external_name in item:
                external_value = item[a.external_name]
                value = a.marshaller.load(external_value)
                kwargs[a.internal_name] = value
        loaded = self.marshalled_type(**kwargs)
        return loaded

    def dump(self, item: T) -> ExternalType:
        if hasattr(item, "get"):

            def getter(obj, key):
                return obj.get(key)

        else:
            getter = getattr

        dumped = {}
        for a in self.attr_configs:
            if a.dump:
                value = getter(item, a.internal_name)
                if value not in a.exclude_dumped_values:
                    external_value = a.marshaller.dump(value)
                    dumped[a.external_name] = external_value
        return dumped



import inspect
from dataclasses import dataclass
from enum import Enum
from typing import Type, Optional

from marshy.marshy_context import MarshyContext
from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.marshaller import marshaller_abc
from marshy.marshaller.enum_marshaller import EnumMarshaller


@dataclass
class EnumMarshallerFactory(MarshallerFactoryABC):
    priority: int = 100
    allow_unknown: bool = False

    def create(
        self, context: MarshyContext, type_: Type
    ) -> Optional[marshaller_abc.MarshallerABC]:
        if inspect.isclass(type_) and issubclass(type_, Enum):
            return EnumMarshaller[type_](type_, self.allow_unknown)



from dataclasses import dataclass
from typing import TypeVar, Iterable, List, Tuple, Union, Type

from marshy.types import ExternalItemType, ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC

T = TypeVar("T")


@dataclass(frozen=True)
class TupleMarshaller(MarshallerABC[Iterable[T]]):
    """
    Marshaller for iterable types (lists)
    """

    marshalled_type: Type
    item_marshallers: Tuple[MarshallerABC[T], ...]

    def load(self, item: Union[List[ExternalType], ExternalItemType]) -> Iterable[T]:
        if isinstance(item, dict):
            loaded = tuple(
                m.load(item[f"t{i}"]) for i, m in enumerate(self.item_marshallers)
            )
            return loaded
        loaded = tuple(m.load(i) for m, i in zip(self.item_marshallers, item))
        return loaded

    def dump(self, item: Iterable[T]) -> List[ExternalType]:
        dumped = [m.dump(i) for m, i in zip(self.item_marshallers, item)]
        return dumped



from abc import ABC, abstractmethod
from typing import Generic, Type, TypeVar

from marshy.types import ExternalType

T = TypeVar("T")


class MarshallerABC(ABC, Generic[T]):
    marshalled_type: Type[T]

    @abstractmethod
    def load(self, item: ExternalType) -> T:
        """Marshall the object given"""

    @abstractmethod
    def dump(self, item: T) -> ExternalType:
        """Unmarshall the object given"""



import importlib
import typing

import typing_inspect

from marshy.errors import MarshallError

_TYPES_BY_ORIGIN = {
    t.__origin__: t for t in typing.__dict__.values() if hasattr(t, "__origin__")
}


def resolve_forward_refs(type_: typing.Type) -> typing.Type:
    origin = typing_inspect.get_origin(type_)
    if origin is not None:
        args = list(resolve_forward_refs(a) for a in typing_inspect.get_args(type_))
        typing_origin = _TYPES_BY_ORIGIN.get(origin)
        origin = typing_origin or origin
        return origin[tuple(args)]
    if not typing_inspect.is_forward_ref(type_):
        return type_
    import_name = typing_inspect.get_forward_arg(type_)
    import_path = import_name.split(".")
    import_module = ".".join(import_path[:-1])
    if not import_module:
        raise MarshallError(f"InvalidForwardRef:{type_}:Use full module name!")
    imported_module = importlib.import_module(import_module)
    return_type = getattr(imported_module, import_path[-1])
    return return_type



from dataclasses import dataclass
from enum import Enum
from typing import TypeVar, Union

from marshy.errors import MarshallError
from marshy.marshaller.marshaller_abc import MarshallerABC

T = TypeVar("T", bound=Enum)
ERROR_PREFIX = "INVALID_VALUE__"


@dataclass(frozen=True)
class EnumMarshaller(MarshallerABC[T]):
    """
    Marshaller for enums
    """

    marshalled_type: T
    allow_unknown: bool = False

    def load(self, item: Union[str, int, float]) -> T:
        try:
            loaded = self.marshalled_type[item]
            return loaded
        except KeyError as e:
            if self.allow_unknown:
                pseudo_member = generate_pseudo_member(self.marshalled_type, item)
                return pseudo_member
            raise MarshallError(e) from e

    def dump(self, item: T) -> Union[str, int, float]:
        dumped = item.name
        return dumped


# pylint: disable=W0212
def generate_pseudo_member(enum: T, value: Union[str, int, float]):
    pseudo_member = object.__new__(enum)
    pseudo_member._name_ = f"{ERROR_PREFIX}{str(value).upper()}"
    pseudo_member._value_ = value
    # noinspection PyProtectedMember
    pseudo_member = enum._value2member_map_.setdefault(value, pseudo_member)
    return pseudo_member



from dataclasses import dataclass
from typing import Type, Optional

from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.marshaller import marshaller_abc
from marshy.marshy_context import MarshyContext

MARSHALLER_FACTORY = "__marshaller_factory__"


@dataclass
class FactoryMarshallerFactory(MarshallerFactoryABC):
    priority: int = 110

    def create(
        self, context: MarshyContext, type_: Type
    ) -> Optional[marshaller_abc.MarshallerABC]:
        factory = getattr(type_, MARSHALLER_FACTORY, None)
        if factory is not None:
            return factory(context)



import dataclasses
import inspect
from typing import Type, Optional, List, get_type_hints, Tuple

from marshy.errors import MarshallError
from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.marshaller import marshaller_abc
from marshy.marshaller.deferred_marshaller import DeferredMarshaller
from marshy.marshaller.obj_marshaller import ObjMarshaller, AttrConfig, attr_config
from marshy.marshaller.property_marshaller import PropertyConfig, PropertyMarshaller
from marshy.marshy_context import MarshyContext
from marshy.utils import resolve_forward_refs


@dataclasses.dataclass
class DataclassMarshallerFactory(MarshallerFactoryABC):
    priority: int = 100
    exclude_dumped_values: Tuple = tuple()

    def create(
        self, context: MarshyContext, type_: Type
    ) -> Optional[marshaller_abc.MarshallerABC]:
        if not dataclasses.is_dataclass(type_):
            return
        return dataclass_marshaller(
            type_, context, exclude_dumped_values=self.exclude_dumped_values
        )


def get_property_configs_for_type(
    type_: Type,
    context: MarshyContext,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    exclude_dumped_values: Tuple = DataclassMarshallerFactory.exclude_dumped_values,
):
    property_configs = []
    for p in inspect.getmembers(type_, lambda o: isinstance(o, property)):
        name = p[0]
        if skip(name, include, exclude):
            continue
        prop = p[1]
        type_hints = get_type_hints(prop.fget)
        if "return" not in type_hints:
            raise MarshallError(f"UnannotatedProperty:{type_}:{name}")
        prop_type = type_hints.get("return")

        prop_type = resolve_forward_refs(prop_type)
        marshaller = DeferredMarshaller[prop_type](prop_type, context)
        property_config = PropertyConfig(
            name,
            marshaller,
            prop,
            load=bool(prop.fset),
            exclude_dumped_values=exclude_dumped_values,
        )
        property_configs.append(property_config)
    return property_configs


def skip(name: str, include: Optional[List[str]], exclude: Optional[List[str]]) -> bool:
    if include is not None and name not in include:
        return True
    if exclude is not None and name in exclude:
        return True
    return False


# pylint: disable=R0913
def get_attr_configs_for_type(
    type_: Type,
    context: MarshyContext,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    exclude_dumped_values: Tuple = (None,),
):
    # noinspection PyDataclass
    fields: List[dataclasses.Field] = dataclasses.fields(type_)
    try:
        types = get_type_hints(type_, globalns=None, localns=None)
        attr_configs = [
            attr_config(
                internal_name=f.name,
                marshaller=DeferredMarshaller(types[f.name], context),
                exclude_dumped_values=exclude_dumped_values,
            )
            for f in fields
            if not skip(f.name, include, exclude)
        ]
    except NameError:
        # Still supporting old style forward refs without futures...
        attr_configs = [
            attr_config(
                internal_name=f.name,
                marshaller=DeferredMarshaller(f.type, context),
                exclude_dumped_values=exclude_dumped_values,
            )
            for f in fields
            if not skip(f.name, include, exclude)
        ]
    return attr_configs


def dataclass_marshaller(
    type_: Type,
    context: MarshyContext,
    custom_attr_configs: Optional[List[AttrConfig]] = None,
    custom_property_configs: Optional[List[PropertyConfig]] = None,
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    exclude_dumped_values: Tuple = (None,),
):
    exclude_list = exclude or []
    if custom_attr_configs:
        exclude_list.extend(a.external_name for a in custom_attr_configs)
    if custom_property_configs:
        exclude_list.extend(p.external_name for p in custom_property_configs)
    attr_configs = get_attr_configs_for_type(
        type_, context, include, exclude, exclude_dumped_values
    )
    property_configs = get_property_configs_for_type(
        type_, context, include, exclude_list, exclude_dumped_values
    )
    attr_configs.extend(custom_attr_configs or [])
    property_configs.extend(custom_property_configs or [])
    marshaller = ObjMarshaller[type_](type_, attr_configs)
    if property_configs:
        marshaller = PropertyMarshaller(marshaller, tuple(property_configs))
    return marshaller



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="marshy",
    author="Tim O'Farrell",
    author_email="tofarr@gmail.com",
    description="A convention over configuration approach to object marshalling.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tofarr/marshy",
    packages=setuptools.find_packages(include=["marshy*"]),
    install_requires=["typing-inspect~=0.7", "injecty~=0.0"],
    python_requires=">=3.7",
    extras_require={
        "dev": [
            "black~=23.3",
            "marshmallow-dataclass~=8.5",
            "pytest~=7.2",
            "pytest-cov~=4.0",
            "pytest-xdist~=3.2",
            "pylint~=3.0",
        ],
    },
    setup_requires=["setuptools-git-versioning"],
    setuptools_git_versioning={"enabled": True, "dirty_template": "{tag}"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from dataclasses import dataclass
from typing import Type, Optional

import typing_inspect

from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.marshaller import marshaller_abc
from marshy.marshaller.deferred_marshaller import DeferredMarshaller
from marshy.marshaller.iterable_marshaller import IterableMarshaller
from marshy.marshy_context import MarshyContext


@dataclass
class ListMarshallerFactory(MarshallerFactoryABC):
    priority: int = 100
    type_name_attr: str = "__type__"

    def create(
        self, context: MarshyContext, type_: Type
    ) -> Optional[marshaller_abc.MarshallerABC]:
        origin = typing_inspect.get_origin(type_)
        if origin in (list, set, frozenset):
            item_type = typing_inspect.get_args(type_)[0]
            item_marshaller = DeferredMarshaller(item_type, context)
            marshaller = IterableMarshaller[item_type](type_, item_marshaller, origin)
            return marshaller



from dataclasses import dataclass
from typing import Type, Optional

import typing_inspect

from marshy.factory.marshaller_factory_abc import MarshallerFactoryABC
from marshy.marshaller import marshaller_abc
from marshy.marshaller.deferred_marshaller import DeferredMarshaller
from marshy.marshaller.iterable_marshaller import IterableMarshaller
from marshy.marshaller.tuple_marshaller import TupleMarshaller
from marshy.marshy_context import MarshyContext


@dataclass
class TupleMarshallerFactory(MarshallerFactoryABC):
    priority: int = 90
    type_name_attr: str = "__type__"

    def create(
        self, context: MarshyContext, type_: Type
    ) -> Optional[marshaller_abc.MarshallerABC]:
        origin = typing_inspect.get_origin(type_)
        if origin is tuple:
            args = typing_inspect.get_args(type_)
            if len(args) == 2 and args[1] is Ellipsis:
                item_marshaller = DeferredMarshaller(args[0], context)
                marshaller = IterableMarshaller[type_](type_, item_marshaller, tuple)
                return marshaller
            marshallers = tuple(DeferredMarshaller(t, context) for t in args)
            return TupleMarshaller[type_](type_, marshallers)



from typing import Union, Dict, List

ExternalType = Union[
    None,
    str,
    bool,
    int,
    float,
    List["marshy.types.ExternalType"],
    "marshy.types.ExternalItemType",
]
ExternalItemType = Dict[str, ExternalType]



from dataclasses import dataclass
from typing import TypeVar, Tuple

from marshy import ExternalType
from marshy.marshaller.marshaller_abc import MarshallerABC
from marshy.types import ExternalItemType

T = TypeVar("T")


@dataclass(frozen=True)
class PropertyConfig:
    external_name: str
    marshaller: MarshallerABC
    prop: property
    load: bool = True
    dump: bool = True
    exclude_dumped_values: Tuple = ()


class PropertyMarshaller(MarshallerABC[T]):
    """
    Marshaller which wraps another and uses property setters / getters.
    """

    # noinspection PyDataclass
    def __init__(
        self, marshaller: MarshallerABC[T], property_configs: Tuple[PropertyConfig]
    ):
        self.marshalled_type = marshaller.marshalled_type
        self.marshaller = marshaller
        self.property_configs = property_configs

    def load(self, item: ExternalItemType) -> T:
        loaded = self.marshaller.load(item)
        for property_config in self.property_configs:
            if property_config.load:
                external_value = item.get(property_config.external_name)
                value = property_config.marshaller.load(external_value)
                property_config.prop.fset(loaded, value)
        return loaded

    def dump(self, item: T) -> ExternalType:
        dumped: ExternalItemType = self.marshaller.dump(item)
        for property_config in self.property_configs:
            if property_config.dump:
                value = property_config.prop.fget(item)
                if value not in property_config.exclude_dumped_values:
                    external_value = property_config.marshaller.dump(value)
                    dumped[property_config.external_name] = external_value
        return dumped



from typing import Type, Any, Optional, TypeVar

from marshy.marshy_context import marshy_context, MarshyContext
from marshy.types import ExternalType

_default_context = None
T = TypeVar("T")


# pylint: disable=W0603
def get_default_marshy_context() -> MarshyContext:
    global _default_context
    if not _default_context:
        _default_context = marshy_context()
    return _default_context


def load(type_: Type[T], to_load: ExternalType) -> T:
    return get_default_marshy_context().load(type_, to_load)


def dump(obj: Any, type_: Optional[Type] = None) -> ExternalType:
    return get_default_marshy_context().dump(obj, type_)
